\   3iArt-butterfly.fs  (GForth code), part of 3-instruction Forth for Arduino (3iForth)
\   Author: Assad Ebrahim - assadebrahim2000@gmail.com
\   Documentation/Publication: www.mathscitech.org/articles/TinyPhoto
\   2021-06-11
\   Copyright (c) 2021, All Rights Reserved.  
\   License is given for users to use/modify the code provided attribution (this header) is preserved, and changes are made available for others (repost to comments of the above site is acceptable).  For 3-instruction Forth for Arduino, contact the author.

\ This program converts 0-1 pixel matrix to vertical byte codes to display images on OLED screens.
\ TODO: automate the total software chain incorporating the manual steps described.  This will allow converting img in < 5 sec. 
\       currently takes c.10 min per image file to produce OLED ready graphic

\ Data format is column major (scan left to right along rows, then top to bottom)
\ Row is padded to multiple of 8 (byte)
\ Columns are padded to multiple of byte for vertical oriented LCD

\ User instructions:
\  1. use IrfanView to resize full color depth image to 128x64 pixels or smaller, and save to BMP
\  2. use online tool Image Binarizer (https://www.dcode.fr/binary-image) to convert to 0-1 matrix:
\        upload image, custom threshold @ 25, set 1=black 0=white
\        threshold determines black or white after conversion of color pixel to greyscale 
\        using Rec 601 luma formula (30% red, 60% green, 10% blue):  Y=0.2989R+0.5870G+0.1140B 
\        paste output into text file
\  3. in notepad++ pad out rows and columns to multiples of byte (use column editor)
\     use macro pixels>b8c, (Ctrl+Shift+Alt+P) to record a macro that converts a byte-padded row.  
\     convert all rows, to create the image_map format ready to load into GForth
\  4. in GForth, load the image_map in memory
\                create image descriptors as Forth constants
\                use 3iArt_rel.fs to extract and create the LCD vertical byte codes ready to transmit to OLED
\                can export/save  LCD vertical byte codes to send image directly and skip this transform step (fast draw in hardware)


\ math
: /up   ( a b -- ceil[a/b]  )   /mod swap 0 > if 1 + then ;  \ integer divide & round up     
: /dn   ( a b -- floor[a/b] )   / ;
: >b8   ( 8-bits -- byte )   \ convert 8 bits (lsb on TOS) to byte.  \ test: 219 .b  1 1 0 1  1 0 1 1  >b8 .
    0 8 1 do or swap i lshift loop or ;  
: nrev  ( a1 a2 a3 .. aN  N -- aN aN-1 .. a3 a2 a1 )  1 do i roll loop ;   \ reverse stack


variable iw  \ image bit-width
variable ih  \ image bit-height
variable pw  \ padding bit-width to 8 bits
variable ph  \ padding bit-height to 8 bits
variable #b  \ #bytes in a command (less than MAX_I2C)

: pw@ pw c@ ;
: ph@ ph c@ ;
: #b@ #b c@ ;
: image-init ( iw ih -- bytes-per-transmission )  \ image width and height in bits
  2dup 
  ih c! iw c!
  8 /up 8 * ph c!  \ height/width padded to next highest byte
  8 /up 8 * pw c!
  2 begin pw@ over /mod #b c! swap 1 + dup rot 0 = swap 4 = or until drop \ find largest multiple of bytes less than MAX_I2C_PACKET
  ;
: lb# ( i j -- jmod lb# )  \ ex: 0 0 -> 0 0;   0 11 -> 3 1;  2 11 -> 3 13 ; 
    >r pw@ * r@ + 8 / r> 8 mod swap ; \ provides linear byte # that holds the desired bit
: pix@+ ( adr --  pix-val  adr' )  \ obtains pixel (or byte) from stored memory map, in downward access
    dup c@ swap pw@ 8 / + ;
: jmask8 ( j -- ) \ mask to pick off jth bit from left (j is 0 indexed, 0 to 7)
    8 1 - swap -   1 swap lshift ;  
: pickbit ( byte j -- bit )  \ form bit mask, apply to byte, and return bit value
    jmask8 and 0 > if 1 else 0 then ;  
: >oled ( image-map i j -- byte-lcd ) \ extract LCD graphic byte and send
    lb# rot + 
    8 0 do pix@+ -rot over pickbit swap rot loop 2drop \  find the column of bits from the respective bytes
    8 nrev >b8   \ create byte of data for OLED/LCD
;
: cmd_max?  ( #ct -- t/f ) #b@ mod 0 = ;
: bmp>oled ( w h image-map -- ) 
    -rot image-init pw@ . ." x " ph@ . cr
    ph@ 8 / 0 do cr i . ." :" cr         \ loop through rows
       pw@ 0 do dup j 8 * i >oled        \ loop through columns
                i cmd_max? if cr then hex.   \  record vertical byte code ready to transmit to OLED
             loop 
       pw@ 128 < if cr then         \ only write newline to OLED if row is less than width of OLED (128 pix)
    loop drop ;
